/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file mainAleParser.cpp
 *
 * @brief Main file when using MAiNGO with the ALE parser.
 *
 **********************************************************************************/

#include "MAiNGO.h"
#include "aleModel.h"
#include "mpiUtilities.h"
#include "programParser.h"

#include "symbol_table.hpp"

#include <memory>

/**
 * @brief Main function managing MAiNGO for standalone use
 *
 * Retrieves settings and calls the branch-and-bound solver
 */
int
main(int argc, char *argv[])
{

#ifdef HAVE_MAiNGO_MPI
    // Initialize MPI and corresponding variable
    MPI_Init(&argc, &argv);
    int _rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
    // Mute cout for workers to avoid multiple outputs
    std::ostringstream mutestream;
    std::streambuf *coutBuf = std::cout.rdbuf();
    std::streambuf *cerrBuf = std::cerr.rdbuf();
    MAiNGO_IF_BAB_WORKER
        std::cout.rdbuf(mutestream.rdbuf());
        std::cerr.rdbuf(mutestream.rdbuf());
    MAiNGO_END_IF
#endif

    // Define model and MAiNGO pointers
    std::shared_ptr<maingo::AleModel> myModel;
    std::shared_ptr<maingo::MAiNGO> myMAiNGO;

    ale::symbol_table symbols;
    try {
        std::string problemFile = "problem.txt";
        if (argc >= 2) {
            problemFile = argv[1];
        }
        std::cout << "Reading problem from file " << problemFile << ".\n";
        std::ifstream input(problemFile);
        if (input.is_open()) {
            maingo::ProgramParser par(input, symbols);
            maingo::Program prog;
            par.parse(prog);
            if (par.fail()) {
                throw std::invalid_argument("  Error: Encountered an error while parsing the problem file");
            }

            myModel  = std::make_shared<maingo::AleModel>(prog, symbols);    // define a model object implemented in ale interface
            myMAiNGO = std::make_shared<maingo::MAiNGO>(myModel);            // define a MAiNGO object with the corresponding model
        }
        else {
            throw std::invalid_argument("  Error: Could not open problem file " + problemFile);
        }
        input.close();
    }
    catch (std::exception &e) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr << std::endl
                      << "  Encountered exception:" << std::endl
                      << e.what() << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }
    catch (...) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr << std::endl
                      << "  Encountered an unknown fatal error during initialization. Terminating." << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }

    // Read settings from file
    std::string settingsFile;
    if (argc >= 3) {
        if (argc > 3) {
            std::cout << "  Warning: Accept only the problem and settings file names as input. Ignoring additional command line arguments." << std::endl
                      << std::endl;
        }
        settingsFile = argv[2];
    }
    myMAiNGO->read_settings(settingsFile);
    //myMAiNGO->set_log_file_name("my_log_file");                                           // Set name of log file, default name is bab
    //myMAiNGO->set_csv_file_name("my_csv_iterations_file","my_csv_general_file");          // Set names of csv files, defaults names are bab_Report_Iterations and bab_Report_General


    // Print MANGO and copyright
    // myMAiNGO->print_MAiNGO(std::cout);

#ifdef HAVE_MAiNGO_MPI
    // Turn on output again
    std::cout.rdbuf(coutBuf);
    std::cerr.rdbuf(cerrBuf);
#endif

    // Solve the problem
    maingo::RETCODE maingoStatus;
    try {
        // Write gams file, default file name is MAiNGO_GAMS_File, default solver is SCIP, default for using min and max is true,
        // default for using hyperbolic functions is true, default for ignoring bounding functions is false, default for writing relaxation-only constraints is true
        // This can also be done by using the MAiNGO Setting writeToOtherLanguage
        MAiNGO_IF_BAB_MANAGER
            // myMAiNGO->parse_maingo_to_other_language(maingo::PARSING_LANGUAGE::ALE,"my_problem_file_MAiNGO","SCIP",/*useMinMax*/true,/*useTrig*/true,/*ignoreBoundingFuncs*/false,/*useRelOnly*/true);
        MAiNGO_END_IF
        maingoStatus = myMAiNGO->solve();
    }
    catch (std::exception &e) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr << std::endl
                      << e.what() << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }
    catch (...) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr << std::endl
                      << "  Encountered an unknown fatal error during solution. Terminating." << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }

    MAiNGO_MPI_FINALIZE return 0;
}