/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file mainCppApi.cpp
 *
 * @brief Main file for using MAiNGO as a standalone solver.
 *
 **********************************************************************************/

#include "01_BasicExample/problem.h"
#include "MAiNGO.h"
#include "mpiUtilities.h"
// #include "02_FlowsheetPowerCycle/problemCaseStudy2LCOE.h"
// #include "03_Biobjective/problemEpsCon.h"

#include <memory>


/**
 * @brief Main function managing MAiNGO for standalone use
 *
 * Retrieves settings and calls the branch-and-bound solver
 */
int
main(int argc, char *argv[])
{

#ifdef HAVE_MAiNGO_MPI
    // Initialize MPI and corresponding variable
    MPI_Init(&argc, &argv);
    int _rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
    // Mute cout for workers to avoid multiple outputs
    std::ostringstream mutestream;
    std::streambuf *coutBuf = std::cout.rdbuf();
    std::streambuf *cerrBuf = std::cerr.rdbuf();
    MAiNGO_IF_BAB_WORKER
        std::cout.rdbuf(mutestream.rdbuf());
        std::cerr.rdbuf(mutestream.rdbuf());
    MAiNGO_END_IF
#endif

    // Define model and MAiNGO objects
    std::shared_ptr<Model> myModel;
    std::unique_ptr<maingo::MAiNGO> myMAiNGO;
    try {
        myModel  = std::make_shared<Model>();                                       // Define a model object implemented in problem.h
        myMAiNGO = std::unique_ptr<maingo::MAiNGO>(new maingo::MAiNGO(myModel));    // Define a MAiNGO object with the corresponding model
    }
    catch (std::exception &e) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr
                << std::endl
                << e.what() << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }
    catch (...) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr
                << std::endl
                << "  Encountered an unknown fatal error during initialization. Terminating." << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }

    // Read settings from file
    std::string settingsFile;
    if (argc >= 2) {
        if (argc > 2) {
            std::cout << "  Warning: Accept only the settings file name as input. Ignoring additional command line arguments." << std::endl
                      << std::endl;
        }
        settingsFile = argv[1];
    }
    myMAiNGO->read_settings(settingsFile);
    //myMAiNGO->set_log_file_name("my_log_file");                                           // Set name of log file, default name is bab
    //myMAiNGO->set_csv_file_name("my_csv_iterations_file","my_csv_general_file");          // Set names of csv files, defaults names are bab_Report_Iterations and bab_Report_General

    // myMAiNGO->print_MAiNGO(std::cout);

#ifdef HAVE_MAiNGO_MPI
    // Turn on output again
    std::cout.rdbuf(coutBuf);
    std::cerr.rdbuf(cerrBuf);
#endif

    // Solve the problem
    maingo::RETCODE maingoStatus;
    try {
        // Write gams file, default file name is MAiNGO_GAMS_File, default solver is SCIP, default for using min and max is true,
        // default for using hyperbolic functions is true, default for ignoring bounding functions is false, default for writing relaxation-only constraints is true
        // This can also be done by using the MAiNGO Setting writeToOtherLanguage
        MAiNGO_IF_BAB_MANAGER
            // myMAiNGO->parse_maingo_to_other_language(maingo::PARSING_LANGUAGE::ALE,"my_problem_file_MAiNGO","SCIP",/*useMinMax*/true,/*useTrig*/true,/*ignoreBoundingFuncs*/false,/*useRelOnly*/true);
        MAiNGO_END_IF
        maingoStatus = myMAiNGO->solve();
        // Use this function instead of solve() for solving bi-objective problems using the epsilon-constraint method (don't forget to include the example problem in problemEpsCon.h):
        // maingoStatus = myMAiNGO->solve_epsilon_constraint();
    }
    catch (std::exception &e) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr
                << std::endl
                << e.what() << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }
    catch (...) {
        MAiNGO_IF_BAB_MANAGER
            std::cerr
                << std::endl
                << "  Encountered an unknown fatal error during solution. Terminating." << std::endl;
        MAiNGO_END_IF
        MAiNGO_MPI_FINALIZE return (-1);
    }

    MAiNGO_MPI_FINALIZE return 0;
}