/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file settings.h
 *
 * @brief File containing declarations of everything needed to specify settings
 *        for MAiNGO.
 *
 **********************************************************************************/

#pragma once

#include "babUtils.h"


/**
*   @namespace maingo
*   @brief namespace holding all essentials of MAiNGO
*/
namespace maingo {


/**
* @enum VERB
* @brief Enum for controlling the output level of solvers (i.e., how much should be printed on the screen and, possibly, to the log file).
*/
enum VERB {
    VERB_NONE = 0, /*!< Print no output whatsoever. */
    VERB_NORMAL,   /*!< For BranchAndBound, this means that regular output on solution progress is printed according to the specified print frequency, as well as when the incumbent is updated. For LowerBoundingSolver and UpperBoundingSolver, only critical output is given (e.g., important warnings). */
    VERB_ALL       /*!< Highest output level (very noisy!). */
};


/**
* @enum OUTSTREAM_VERB
* @brief Enum for controlling the output and writing level of MAiNGO (i.e., where should the output be written to, outstream and/or log)
*/
enum OUTSTREAM_VERB {
    OUTSTREAM_NONE = 0,  /*!< Print and write no output whatsoever. */
    OUTSTREAM_OUTSTREAM, /*!< Print only to output stream (standard ist std::cout) */
    OUTSTREAM_LOG,       /*!< Write to log only */
    OUTSTREAM_BOTH       /*!< Print to output stream AND write the same information to log */
};


/**
* @enum PARSING_LANGUAGE
* @brief Enum for representing possible modeling language for parsing a MAiNGO model to.
*/
enum PARSING_LANGUAGE {
    LANG_NONE = -1,    //!< A different parsing language is not set
    AMPL,              //!< AMPL
    ALE,               //!< ALE
    BARON,             //!< BARON
    GAMS,              //!< GAMS
    PYOMO,             //!< PYOMO
    NLP_LANG           //!< NLP
};


/**
*   @namespace maingo::lbp
*   @brief namespace holding all essentials of the lower bounding solver
*/
namespace lbp {


/**
    * @enum LBP_SOLVER
    * @brief Enum for selecting the STRATegy so be used for solving the lower bounding problems
    */
enum LBP_SOLVER {
    SOLVER_MAiNGO   = 0, /*!< (=0): MAiNGO intern lower bounding solver consisting of linearizing the objective function at only 1 point and minimizing the linearization over box constraints */
    SOLVER_INTERVAL = 1, /*!< (=1): solution of lower bounding problems using only interval based relaxations */
    SOLVER_CPLEX    = 2, /*!< (=2): solution of lower bounding linear programs using CPLEX   */
    SOLVER_CLP      = 3  /*!< (=3): solution of lower bounding linear programs using CLP   */
};

/**
    * @enum OBBT
    * @brief Enum for communicating whether Optimization-Based Bound Tightening should consider only feasibility or also optimality
    */
enum OBBT {
    OBBT_FEAS = 0, /*!< Consider feasibility only, i.e., maximize and minimize each variable only subject to the relaxed (and linearized) constraints */
    OBBT_FEASOPT   /*!< Consider both feasibility and optimality, i.e., including the objective function cut f_cv<=UBD */
};

/**
    * @enum LINP
    * @brief Enum for selecting the Linearization Points to be used in constructing affine relaxations.
    */
enum LINP {
    LINP_MID = 0,       /*!< (=0) : Linearize only at the midpoint of the current node */
    LINP_INCUMBENT,     /*!< (=1) : Linearize at the incumbent value if it is in the current interval, else linearize at mid point, if using the subgradient interval heuristic, the heuristic also linearizes each operation at the incumbent if possible */
    LINP_KELLEY,        /*!< (=2) : Linearize at points determined via an adapted version of Kelley's algorithm, each function is treated individually */
    LINP_SIMPLEX,       /*!< (=3) : Linearize at mid point + (n+1)/2 points given as vertices of the (n+1) simplex where n is the dimension of the problem */
    LINP_RANDOM,        /*!< (=4) : Linearize at mid point + (n+1)/2 random points */
    LINP_KELLEY_SIMPLEX /*!< (=5) : Linearize at mid point + (n+1)/2 points given as vertices of the (n+1) simplex where n is the dimension of the problem and then apply Kelleys algorithm */
};

}    // end namespace lbp


/**
*   @namespace maingo::ubp
*   @brief namespace holding all essentials of the upper bounding solvers
*/
namespace ubp {

/**
    * @enum UBP_SOLVER
    * @brief Enum for selecting the STRATegy so be used for solving the upper bounding problems
    */
enum UBP_SOLVER {
    SOLVER_EVAL = 0,   /*!< (=0): no optimization, simple function evaluation at solution point of LBP  */
    SOLVER_COBYLA,     /*!< (=1): local optimization using COBYLA (derivative free solver within NLopt, uses linear approximations via simplex of nvar+1 points) */
    SOLVER_BOBYQA,     /*!< (=2): local optimization using BOBYQA (derivative free unconstrained solver in NLopt, constructs quadratic approximations; constraints are moved to the objective via augmented Lagrangian method) */
    SOLVER_LBFGS,      /*!< (=3): local optimization using LBFGS (lower-storage BFGS algorithm (i.e., gradient-based) for unconstrained optimization within NLopt; constraints are moved to the objective via augmented Lagrangian method) */
    SOLVER_SLSQP,      /*!< (=4): local optimization using SLSQP (SQP solver within NLopt) */
    SOLVER_IPOPT,      /*!< (=5): local optimization using Ipopt (using the exact Hessian for problems with at most 50 variables, else using L-BFGS) */
    SOLVER_KNITRO,     /*!< (=6): local optimization using Knitro (using the exact Hessian) */
    SOLVER_CPLEX = 42, /*!< (=42): optimization using CPLEX. Called only for (MI)LPs and (MI)QCPs. */
    SOLVER_CLP         /*!< (=43): optimization using CLP. Called only for LPs. */
};


}    // end namespace ubp


/**
* @struct Settings
* @brief Struct for storing settings for MAiNGO
*
* Contains settings for MAiNGO. The default values an be found in settings.cpp.
*
*/
struct Settings {

    /**
    * @brief Constructor, sets default values (cf. settings.cpp).
    */
    Settings();

    /**
    * @brief Default copy constructor.
    */
    Settings(const Settings&) = default;

    /**
    * @brief Default copy assignment.
    */
    Settings& operator=(const Settings&) = default;

    /**
    * @name Tolerances and termination
    */
    /**@{*/
    double epsilonA;      //!< Absolute optimality tolerance, i.e., termination when (UBD-LBD) < BAB_epsilon_a
    double epsilonR;      //!< Relative optimality tolerance, i.e., termination when (UBD-LBD) < BAB_epsilon_r * UBD
    double deltaIneq;     //!< Absolute feasibility tolerance for inequality constraints (i.e., constraint is considered satisfied if gi_(x)<=UBP_delta_ineq
    double deltaEq;       //!< Absolute feasibility tolerance for equality constraints (i.e., constraint is considered satisfied if |hi_(x)|<=UBP_delta_eq
    double relNodeTol;    //!< Relative tolerance for minimum node size

    unsigned BAB_maxNodes;            //!< Maximum number of nodes (i.e., solver terminates when more than BAB_maxnodes are held in memory; used to avoid excessive branching)
    unsigned BAB_maxIterations;       //!< Maximum number of iterations (i.e., maximum number of nodes visited in the Branch-and-Bound tree)
    unsigned maxTime;                 //!< CPU time limit in seconds
    bool confirmTermination;          //!< Whether to ask the user before terminating when reaching time, node, or iteration limits
    bool terminateOnFeasiblePoint;    //!< Whether to terminate as soon as the first feasible point was found (no guarantee of global or local optimality!)
    double targetLowerBound;          //!< Target value for the lower bound on the optimal objective. MAiNGO terminates once LBD>=targetLowerBound  (no guarantee of global or local optimality!)
    double targetUpperBound;          //!< Target value for the upper bound on the optimal objective. MAiNGO terminates once UBD<=targetUpperBound  (no guarantee of global or local optimality!)
    double infinity;                  //!< User definition of infinity (used to initialize UBD and LBD) [currently cannot be set by the user via set_option]
    /**@}*/

    /**
    * @name Pre-processing
    */
    /**@{*/
    unsigned PRE_maxLocalSearches;    //!< Number of local searches in the multistart heuristic during preprocessing at the root node
    unsigned PRE_obbtMaxRounds;       //!< Maximum number of rounds of optimization-based range reduction (OBBT; cf., e.g., Gleixner et al., J. Glob. Optim. 67 (2017) 731; maximizing and minimizing each variable subject to relaxed constraints) at the root node. If >=1 and a feasible point is found during multistart, one round of OBBT using an objective cut (f_cv<=UBD) is conducted as well.
    bool PRE_pureMultistart;          //!< Whether to perform a multistart only. A B&B tree will not be constructed and no lower bounding problems will be solved
    /**@}*/

    /**
    * @name B&B Settings - Tree management & range reduction
    */
    /**@{*/
    babBase::enums::NS BAB_nodeSelection;     //!< How to select the next node to process. See documentation of babBase::enums::NS for possible values.
    babBase::enums::BV BAB_branchVariable;    //!< Which dimension to branch in for the current node. See documentation of babBase::enums::BV for possible values.
    bool BAB_alwaysSolveObbt;                 //!< Whether to solve OBBT (feasibility- and, once a feasible point has been found, also optimality-based) at every BaB node
    bool BAB_dbbt;                            //!< Whether to do a single round of duality based bound tightening (DBBT, cf. Ryoo&Sahinidis, Comput. Chem. Eng. 19 (1995) 551). If false, no DBBT is used. If true, multipliers from CPLEX are used to tighten bounds (essentially for free). we tried additional rounds but without reasonable improvement.
    bool BAB_probing;                         //!< Whether to do probing (cf. Ryoo&Sahinidis, Comput. Chem. Eng. 19 (1995) 551) at every node (can only be done if BAB_DBBT_maxrounds>=1)
    bool BAB_constraintPropagation;           //!< Whether to do constraint propagation. If false, no constraint propagation is executed.
    /**@}*/

    /**
    * @name LBP Settings
    */
    /**@{*/
    lbp::LBP_SOLVER LBP_solver;              //!< Solver for solution of lower bounding problems.
    lbp::LINP LBP_linPoints;                 //!< At which points to linearize for affine relaxation. See documentation of lbp::LINP for possible values.
    bool LBP_subgradientIntervals;           //!< Whether to use the heuristic to improve McCormick relaxations by tightening the range of each factor with the use of subgradients (cf. Najman & Mitsos, JOGO 2019)
    double LBP_obbtMinImprovement;           //!< How much improvement needs to be achievable (relative to initial diameter) to conduct OBBT for a variable
    unsigned LBP_activateMoreScaling;        //!< Number of consecutive iterations without LBD improvement needed to activate more aggressive scaling in LP solver (e.g., CPLEX)
    bool LBP_addAuxiliaryVars;               //!< Whether to add auxiliary variables for common factors in the lower bounding DAG/problem
    unsigned LBP_minFactorsForAux;           //!< Minimum number of common factors to add an auxiliary variable
    unsigned LBP_maxNumberOfAddedFactors;    //!< Maximum number of added factor as auxiliaries
    /**@}*/

    /**
    * @name MC++ settings
    */
    /**@{*/
    bool MC_mvcompUse;      //!< Whether to use multivariate composition theorem for computing McCormick relaxations  (see MC++ documentation for details)
    double MC_mvcompTol;    //!< (see MC++ documentation for details)
    double MC_envelTol;     //!< (see MC++ documentation for details)
    /**@}*/

    /**
    * @name UBP Settings
    */
    /**@{*/
    ubp::UBP_SOLVER UBP_solverPreprocessing;    //!< Solver to be used during pre-processing (i.e., multistart). See documentation of ubp::UBP_SOLVER for possible values.
    unsigned UBP_maxStepsPreprocessing;         //!< Maximum number of steps the local solver is allowed to take in each local run during multistart in pre-processing.
    double UBP_maxTimePreprocessing;            //!< Maximum CPU time the local solver is allowed to take in each local run during multistart in pre-processing. Usually, this should only be a fall-back option to prevent truly getting stuck in local solution.
    ubp::UBP_SOLVER UBP_solverBab;              //!< Solver to be used during Branch-and-Bound. See documentation of ubp::UBP_SOLVER for possible values.
    unsigned UBP_maxStepsBab;                   //!< Maximum number of steps the local solver is allowed to take at each BaB node.
    double UBP_maxTimeBab;                      //!< Maximum CPU time the local solver is allowed to take at each BaB node. Usually, this should only be a fall-back option to prevent truly getting stuck in local solution.
    bool UBP_ignoreNodeBounds;                  //!< Flag indicating whether the UBP solvers should ignore the box constraints of the current node during the B&B (and consider only the ones of the root node instead).
    /**@}*/

    /**
    * @name Epsilon-constraint settings
    */
    /**@{*/
    unsigned EC_nPoints;    //!< Number of points on the Pareto front to be computed in epsilon-constraint method (only available via the C++ API)
    /**@}*/

    /**
    * @name Output
    */
    /**@{*/
    VERB BAB_verbosity;                       //!< How much output to print from Branch & Bound solver. Possible values are VERB_NONE (=0), VERB_NORMAL (=1), VERB_ALL (=2)
    VERB LBP_verbosity;                       //!< How much output to print from Lower Bounding Solver. Possible values are VERB_NONE (=0), VERB_NORMAL (=1), VERB_ALL (=2)
    VERB UBP_verbosity;                       //!< How much output to print from Upper Bounding Solver. Possible values are VERB_NONE (=0), VERB_NORMAL (=1), VERB_ALL (=2)
    unsigned BAB_printFreq;                   //!< After how many iterations to print progress on screen (additionally, a line is printed when a new incumbent is found)
    unsigned BAB_logFreq;                     //!< Like BAB_printFreq, but for log
    OUTSTREAM_VERB outstreamVerbosity;        //!< Where to print or write the output. Possible values are OUTSTREAM_NONE (=0), OUTSTREAM_OUTSTREAM (=1), OUTSTREAM_LOG (=2),  OUTSTREAM_BOTH (=3)
    bool writeLog;                            //!< Whether to write a log file (named bab.log)
    unsigned writeToLogSec;                   //!< Write to log file after a given ammount of CPU seconds
    bool writeResFile;                        //!< Whether to write an additional file containing non-standard information about the solved model
    bool writeCsv;                            //!< Whether to write a csv-log file (named bab.csv). Currently, this only include time, LBD, UBD, and final output
    bool PRE_printEveryLocalSearch;           //!< Whether to print every run during multistart at the root node
    PARSING_LANGUAGE writeToOtherLanguage;    //!< Write to a file in a different modeling language
    /**@}*/
};


}    // end namespace maingo