/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file MAiNGOgetterFunctions.cpp
 *
 * @brief File containing implementation of MAiNGO getter functions.
 *
 **********************************************************************************/

#include "MAiNGO.h"
#include "bab.h"
#include "exceptions.h"


using namespace maingo;


/////////////////////////////////////////////////////////////////////////
// returns the value of the objective at solution point
double
MAiNGO::get_objective_value() const
{
    if ((_maingoStatus != GLOBALLY_OPTIMAL) && (_maingoStatus != FEASIBLE_POINT)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying objective value. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    return _solutionValue;
}


/////////////////////////////////////////////////////////////////////////
// returns the solution point
std::vector<double>
MAiNGO::get_solution_point() const
{
    if (_solutionPoint.empty()) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying solution point. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    std::vector<double> solutionPoint;
    unsigned removed = 0;
    for (unsigned i = 0; i < _nvarOriginal; ++i) {
        if (_removedVariables[i]) {
            // If the variable has been removed from the optimization problem, simply return the middle point of the original interval
            solutionPoint.push_back((_originalVariables[i].get_lower_bound() + _originalVariables[i].get_upper_bound()) / 2.0);
            removed++;
        }
        else {
            // Otherwise simply return the value of the variable at solution point
            solutionPoint.push_back(_solutionPoint[i - removed]);
        }
    }
    return solutionPoint;
}


/////////////////////////////////////////////////////////////////////////
// returns the solution time
double
MAiNGO::get_cpu_solution_time() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying solution time. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    return _solutionTime;
}


/////////////////////////////////////////////////////////////////////////
// returns the solution time
double
MAiNGO::get_wallclock_solution_time() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying solution time. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    return _solutionTimeWallClock;
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the number of iterations
double
MAiNGO::get_iterations() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying number of iterations. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_iterations();
    }
    else {
        return 0;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the maximum number of nodes in memory
double
MAiNGO::get_max_nodes_in_memory() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying number of nodes in memory. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_max_nodes_in_memory();
    }
    else {
        return 1;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning number of UBD problems solved
double
MAiNGO::get_UBP_count() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying UBP count. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_UBP_count();
    }
    else {
        return 1;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning number of LBD problems solved
double
MAiNGO::get_LBP_count() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying LBP count. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_LBP_count();
    }
    else {
        return 0;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the final LBD
double
MAiNGO::get_final_LBD() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying final LBD. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_final_LBD();
    }
    else {
        return _solutionValue;    // In case of an LP, MIP, QP, or MIQP, we take the solution to be exact for now...
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the final absolute gap
double
MAiNGO::get_final_abs_gap() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying final absolute gap. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_final_abs_gap();
    }
    else {
        return 0;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the final relative gap
double
MAiNGO::get_final_rel_gap() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying final relative gap. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    if (_myBaB) {
        return _myBaB->get_final_rel_gap();
    }
    else {
        return 0;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the current MAiNGO status
RETCODE
MAiNGO::get_status() const
{
    if ((_maingoStatus == NOT_SOLVED_YET)) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying final relative gap. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }
    return _maingoStatus;
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the objective and all constraints at solution point
std::vector<double>
MAiNGO::get_model_at_solution_point()
{

    // Problem has to be solved
    if (_solutionPoint.empty()) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying model outputs in get_model_at_solution_point. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }

    std::vector<double> result(_DAGfunctions.size());
    _get_model_at_point(_solutionPoint, result);
    return result;
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the additional model outputs at the solution point
std::vector<std::pair<std::string, double>>
MAiNGO::get_additional_outputs_at_solution_point()
{

    // Return variable additional model outputs
    if (_solutionPoint.empty()) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error querying additional model outputs in get_additional_outputs_at_solution_point. MAiNGO status: " << _maingoStatus;
        throw(MAiNGOException(errmsg.str()));
    }

    std::vector<std::pair<std::string, double>> additionalModelOutputs;
    if ((!_solutionPoint.empty()) && (_noutputVariables > 0 || _nconstantOutputVariables > 0) && (_constantConstraintsFeasible)) {
        return _get_additional_outputs(_solutionPoint);
    }

    return additionalModelOutputs;
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the objective and all constraints at a user specified point
bool
MAiNGO::get_model_at_point(const std::vector<double> &point, std::vector<double> &returnResult)
{
    // DAG has to be constructed
    if (!_DAGconstructed) {
        _construct_DAG();
    }

    if (!_readyToSolve) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error in get_model_at_point. Model has not been set yet.";
        throw(MAiNGOException(errmsg.str()));
    }

    if (point.size() != _nvarOriginal) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: The dimension of the point in function get_model_at_point does not match the dimensions of the set MAiNGO model.";
        throw(MAiNGOException(errmsg.str()));
    }

    // Match each DAG variable to a double variable and skip removed variables
    // Evaluate DAG at point
    std::vector<double> pointUsed;
    for (unsigned int i = 0; i < _nvarOriginal; i++) {
        if (!_removedVariables[i]) {
            pointUsed.push_back(point[i]);
        }
    }

    return _get_model_at_point(pointUsed, returnResult);
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the additional model outputs at a user specified point
std::vector<std::pair<std::string, double>>
MAiNGO::get_additional_outputs_at_point(const std::vector<double> &point)
{
    // DAG has to be constructed
    if (!_DAGconstructed) {
        _construct_DAG();
    }

    if (!_readyToSolve) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: Error in get_additional_output_at_point. Model has not been set yet.";
        throw(MAiNGOException(errmsg.str()));
    }

    if (point.size() != _nvarOriginal) {
        std::ostringstream errmsg;
        errmsg << "  MAiNGO: The dimension of the point in function get_additional_output_at_point does not match the dimensions of the set MAiNGO model.";
        throw(MAiNGOException(errmsg.str()));
    }

    // Match each DAG variable to a double variable and skip removed variables
    // Evaluate DAG at point
    std::vector<double> pointUsed;
    for (unsigned int i = 0; i < _nvarOriginal; i++) {
        if (!_removedVariables[i]) {
            pointUsed.push_back(point[i]);
        }
    }

    return _get_additional_outputs(pointUsed);
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the additional model outputs at a user specified point
std::vector<std::pair<std::string, double>>
MAiNGO::_get_additional_outputs(const std::vector<double> &pointUsed)
{

    // Match each DAG variable to a double variable
    // Evaluate DAG to compute outputs
    std::vector<double> outputRes(_DAGoutputFunctions.size());
    try {
        _DAG.eval(_DAGoutputFunctions.size(), _DAGoutputFunctions.data(), outputRes.data(), _nvar, _DAGvars.data(), pointUsed.data());
    }
    catch (std::exception &e) {
        throw MAiNGOException("  MAiNGO: Error while evaluating additional user outputs.", e);
    }
    catch (...) {
        throw MAiNGOException("  MAiNGO: Error while evaluating additional user outputs.");
    }
    // Get additional output, don't forget the constant ones
    std::vector<std::pair<std::string, double>> additionalModelOutputs(_noutputVariables + _nconstantOutputVariables);
    // Non-constant outputs first
    for (size_t i = 0; i < _noutputVariables; i++) {
        additionalModelOutputs[_nonconstantOutputs[i].indexOriginal] = std::make_pair(_nonconstantOutputs[i].name, outputRes[_nonconstantOutputs[i].indexNonconstant]);
    }
    // Constant outputs second
    for (size_t i = 0; i < _nconstantOutputVariables; i++) {
        additionalModelOutputs[_constantOutputs[i].indexOriginal] = std::make_pair(_constantOutputs[i].name, _constantOutputs[i].constantValue);
    }

    return additionalModelOutputs;
}


////////////////////////////////////////////////////////////////////////////////////////
// function returning the objective and all constraints at a user specified point
bool
MAiNGO::_get_model_at_point(const std::vector<double> &pointUsed, std::vector<double> &returnResult)
{

    std::vector<double> result(_DAGfunctions.size());
    _DAG.eval(_DAGfunctions.size(), _DAGfunctions.data(), result.data(), _nvar, _DAGvars.data(), pointUsed.data());

    bool isFeasible = true;
    // Don't forget the constant functions
    returnResult.clear();    // The vector to be returned
    returnResult.resize(1 + _nineq + _nconstantIneq + _neq + _nconstantEq + _nineqRelaxationOnly + _nconstantIneqRelOnly + _neqRelaxationOnly + _nconstantEqRelOnly + _nineqSquash + _nconstantIneqSquash);
    returnResult[0] = result[0];    // Objective first
    // Non-constant constraints first
    unsigned offset = _nauxiliaryRelOnlyEqs;    // Don't count auxiliary equalities
    for (size_t i = 0; i < _nonconstantConstraints.size() - offset; i++) {
        double value = result[_nonconstantConstraints[i].indexNonconstant];
        switch (_nonconstantConstraints[i].type) {
            case INEQ:
            case INEQ_REL_ONLY: {
                if (value > _maingoSettings.deltaIneq) {
                    isFeasible = false;
                }
                break;
            }
            case EQ:
            case EQ_REL_ONLY: {
                if (std::fabs(value) > _maingoSettings.deltaEq) {
                    isFeasible = false;
                }
                break;
            }
            case INEQ_SQUASH: {
                if (value > 0) {
                    isFeasible = false;
                }
                break;
            }
            default:
                break;
        }
        returnResult[_nonconstantConstraints[i].indexOriginal] = value;
    }
    // Constant constraints second
    for (size_t i = 0; i < _constantConstraints.size(); i++) {
        double value = _constantConstraints[i].constantValue;
        switch (_constantConstraints[i].type) {
            case INEQ:
            case INEQ_REL_ONLY: {
                if (value > _maingoSettings.deltaIneq) {
                    isFeasible = false;
                }
                break;
            }
            case EQ:
            case EQ_REL_ONLY: {
                if (std::fabs(value) > _maingoSettings.deltaEq) {
                    isFeasible = false;
                }
                break;
            }
            case INEQ_SQUASH: {
                if (value > 0) {
                    isFeasible = false;
                }
                break;
            }
            default:
                break;
        }
        returnResult[_constantConstraints[i].indexOriginal] = value;
    }

    return isFeasible;
}