/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file lbpFactory.cpp
 *
 * @brief File containing implementation of factory function for initializing
 *        different lower bounding solver wrappers.
 *
 **********************************************************************************/

#include "exceptions.h"
#include "lbp.h"
#include "lbpClp.h"
#include "lbpInterval.h"

#ifdef HAVE_CPLEX    // If Cmake has found CPLEX this pre-processor variable is set
#include "lbpCplex.h"
#endif


using namespace maingo;
using namespace lbp;


/////////////////////////////////////////////////////////////////////////
// function for initializing the different lower bounding solver wrappers
std::shared_ptr<LowerBoundingSolver>
lbp::make_lbp_solver(mc::FFGraph &DAG, const std::vector<mc::FFVar> &DAGvars, const std::vector<mc::FFVar> &DAGfunctions,
                     const std::vector<babBase::OptimizationVariable> &variables, const unsigned nineqIn, const unsigned neqIn,
                     const unsigned nineqRelaxationOnlyIn, const unsigned neqRelaxationOnlyIn, const unsigned nineqSquashIn,
                     Settings *settingsIn, Logger *loggerIn, std::vector<Constraint> *constraintPropertiesIn)
{

    switch (settingsIn->LBP_solver) {
        case SOLVER_MAiNGO: {
            return std::make_shared<LowerBoundingSolver>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn,
                                                         nineqRelaxationOnlyIn, neqRelaxationOnlyIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn);
        }
        case SOLVER_INTERVAL: {
            return std::make_shared<LbpInterval>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn,
                                                 nineqRelaxationOnlyIn, neqRelaxationOnlyIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn);
        }
        case SOLVER_CPLEX: {
#ifdef HAVE_CPLEX
            return std::make_shared<LbpCplex>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn,
                                              nineqRelaxationOnlyIn, neqRelaxationOnlyIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn);
#else
            // If MAiNGO does not have CPLEX, we use CLP instead
            settingsIn->LBP_solver = SOLVER_CLP;
            loggerIn->print_message("        The option LBP_solver was set to 2 (CPLEX) but your MAiNGO build does not contain CPLEX. Setting it to 3 (CLP).\n",
                                    settingsIn->writeLog, settingsIn->LBP_verbosity, VERB_NONE, settingsIn->outstreamVerbosity);    // Save information that a change was made
                                                                                                                                    // Falling through to CLP
#endif
        }
        case SOLVER_CLP: {
            return std::make_shared<LbpClp>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn,
                                            nineqRelaxationOnlyIn, neqRelaxationOnlyIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn);
        }
        default:
        {
            std::ostringstream errmsg;
            errmsg << "  Error in LbpFactory: Unknown lower bounding solver: " << settingsIn->LBP_solver << std::endl;
            throw(MAiNGOException(errmsg.str()));
        }
    }
}