/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file logger.cpp
 *
 * @brief File implementing the logger class which manages the printing of
 *        output and writing of files.
 *
 **********************************************************************************/

#include "logger.h"

#include <fstream>
#include <sstream>


using namespace maingo;


/////////////////////////////////////////////////////////////////////////
// writes a message to outstream and possibly bablog
void
Logger::print_message(const std::string& message, bool writeLog, VERB verbosityGiven, VERB verbosityNeeded, OUTSTREAM_VERB givenOutstreamVerbosity)
{

    switch (givenOutstreamVerbosity) {

        case OUTSTREAM_OUTSTREAM:
            // Print to _outStream only
            if (verbosityGiven >= verbosityNeeded) {
                (*_outStream) << message << std::flush;
            }
            break;
        case OUTSTREAM_LOG:
            // Save message in log queue to be written later
            if (writeLog && verbosityGiven >= verbosityNeeded) {
                babLine.push(message);
            }
            break;
        case OUTSTREAM_BOTH:
            // Print and write
            if (verbosityGiven >= verbosityNeeded) {
                (*_outStream) << message << std::flush;
                if (writeLog) {
                    babLine.push(message);
                }
            }
            break;
        case OUTSTREAM_NONE:
        default:
            // Don't print or write
            break;
    }
}


/////////////////////////////////////////////////////////////////////////
// writes a message to outstream only without asking for vorbosities
void
Logger::print_message_to_stream_only(const std::string& message, OUTSTREAM_VERB givenOutstreamVerbosity)
{
    if (givenOutstreamVerbosity == OUTSTREAM_BOTH || givenOutstreamVerbosity == OUTSTREAM_OUTSTREAM) {
        (*_outStream) << message << std::flush;
    }
}


/////////////////////////////////////////////////////////////////////////
// sets output stream
void
Logger::set_output_stream(std::ostream* outputStream)
{
    _outStream = outputStream;
}


/////////////////////////////////////////////////////////////////////////
// creates the log file
void
Logger::create_log_file(bool writeLog, OUTSTREAM_VERB givenOutstreamVerbosity)
{
    if (writeLog && (givenOutstreamVerbosity == OUTSTREAM_BOTH || givenOutstreamVerbosity == OUTSTREAM_LOG)) {
        std::ofstream logFile;
        logFile.open(logFileName, std::ios::out);
        logFile.close();
    }
}


/////////////////////////////////////////////////////////////////////////
// creates the log file
void
Logger::create_csv_files(bool pureMultistart, bool printEveryLocalSearch, unsigned int maxLocalSearches, const std::vector<double>& objectivesAtRoot, const std::vector<SUBSOLVER_RETCODE>& feasibleAtRoot)
{

    std::ofstream outfileIterations(csvIterationsName);
    outfileIterations << " Iters,"
#ifdef MAiNGO_DEBUG_MODE
                      << " NodeId,"
                      << " NodeLBD,"
#endif
                      << " LBD, "
                      << " UBD,"
                      << " NodesLeft,"
                      << " AbsGap,"
                      << " RelGap,"
                      << " CPU" << std::endl;
    outfileIterations.close();

    std::ofstream outfileGeneral(csvGeneralName);
    if (pureMultistart) {
        outfileGeneral << "  Pure Multistart " << std::endl;
        outfileGeneral << "  No of local searches," << maxLocalSearches << std::endl;
    }
    if (printEveryLocalSearch) {
        for (unsigned i = 0; i < maxLocalSearches; i++) {
            if (feasibleAtRoot[i] == SUBSOLVER_FEASIBLE) {
                outfileGeneral << "  \tRun No," << i + 1 << ",objective value," << objectivesAtRoot[i] << std::endl;
            }
            else {
                outfileGeneral << "  \tRun No," << i + 1 << ",No feasible point found" << std::endl;
            }
        }
    }
    outfileGeneral.close();
}


/////////////////////////////////////////////////////////////////////////
// writes all lines currently stored in babLine to logFile
void
Logger::write_all_lines_to_log(const std::string& errmsg)
{
    std::ofstream logFile;
    logFile.open(logFileName, std::ios::app);
    while (babLine.size() > 0) {
        logFile << babLine.front();
        babLine.pop();
    }
    if (!errmsg.empty()) {
        logFile << errmsg << std::endl;
    }
    logFile.close();
}


/////////////////////////////////////////////////////////////////////////
// writes all b&B iterations currently stored in babCsv to csv iterations file
void
Logger::write_all_iterations_to_csv(const std::string& errmsg)
{
    std::ofstream csvFile;
    csvFile.open(csvIterationsName, std::ios::app);
    while (babLineCsv.size() > 0) {
        csvFile << babLineCsv.front();
        babLineCsv.pop();
    }
    if (!errmsg.empty()) {
        csvFile << errmsg << std::endl;
    }
    csvFile.close();
}


/////////////////////////////////////////////////////////////////////////
// saves a proper string when a user wants to read in a setting file
void
Logger::save_settings_file_name(const std::string& fileName, bool fileFound)
{

    // User wants to read in a file
    _nSettingFiles++;
    int mapNumber   = -1 * _nSettingFiles;
    std::string str = "";
    if (fileFound) {
        // If file has been found, generate string
        str = "\n  Read settings from file " + fileName + ".";
    }
    else {
        // If file has not been found generate a different string
        str = "\n  Warning: Could not open settings file " + fileName + ". Proceeding with default settings.\n           For further information regarding settings, refer to Readme.md.";
    }
    // Then insert this string at a proper position in the map
    _userSetSettings[mapNumber] = str;
}


/////////////////////////////////////////////////////////////////////////
// save a user-set setting in map
void
Logger::save_setting(SETTING_NAMES settingName, const std::string& str)
{

    switch (settingName) {
        case UNKNOWN_SETTING: {
            int wrongSettings = static_cast<int>(settingName);
            while (_userSetSettings.find(wrongSettings) != _userSetSettings.end()) {
                wrongSettings++;
            }
            _userSetSettings[wrongSettings] = str;
        } break;
        default:
            // Replace/insert the new setting string
            _userSetSettings[static_cast<int>(settingName)] = str;
            break;
    }
}


/////////////////////////////////////////////////////////////////////////
// print and or write user-set settings
void
Logger::print_settings(bool writeLog, VERB verbosityGiven, VERB verbosityNeeded, OUTSTREAM_VERB givenOutstreamVerbosity)
{

    // First check if any setting was changed at all
    if (!_userSetSettings.empty()) {
        bool someSettingChanged = (_userSetSettings.rbegin()->first > 0);    // This checks if there is at least one entry with a positive key, since these belong to the actual settings


        // If so, we print two additional lines two frame the actual settings (otherwise, we only give output about potential read attempts to empty or non-existing settings files
        if (someSettingChanged) {
            _userSetSettings[0] = "Settings set by the user:";
        }
        std::string str = "";
        for (std::map<int, std::string>::iterator it = _userSetSettings.begin(); it != _userSetSettings.end(); ++it) {
            if (it->first > 0) {
                str += "    " + (it->second) + "\n";
            }
            else {
                str += "  " + (it->second) + "\n";
            }
        }
        if (someSettingChanged) {
            str += "  Done.\n";
        }
        print_message(str, writeLog, verbosityGiven, verbosityNeeded, givenOutstreamVerbosity);
    }
}


/////////////////////////////////////////////////////////////////////////
// clear logging information
void
Logger::clear()
{
    babLine            = std::queue<std::string>();
    babLineCsv         = std::queue<std::string>();
    reachedMinNodeSize = false;
    if (logFileName.empty()) {
        logFileName = "bab.log";
    }
    if (csvIterationsName.empty()) {
        csvIterationsName = "bab_Report_Iterations.csv";
    }
    if (csvGeneralName.empty()) {
        csvGeneralName = "bab_Report_General.csv";
    }
    if (resFileName.empty()) {
        resFileName = "MAiNGO_res.txt";
    }
}