/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * @file ubpFactory.cpp
 *
 * @brief File containing implementation of factory function for initializing 
 *        different upper bounding solver wrappers.
 *
 **********************************************************************************/

#include "exceptions.h"
#include "ubp.h"
#include "ubpClp.h"
#include "ubpIpopt.h"
#include "ubpNLopt.h"

#ifdef HAVE_CPLEX    // If CMake has found CPLEX ,this pre-processor variable is defined
#include "ubpCplex.h"
#endif

#ifdef HAVE_KNITRO    // If CMake has found KNITRO, this pre-processor variable is defined
#include "ubpKnitro.h"
#endif


using namespace maingo;
using namespace ubp;


/////////////////////////////////////////////////////////////////////////
// Function for initializing the different upper bounding solver wrappers
std::shared_ptr<UpperBoundingSolver>
ubp::make_ubp_solver(mc::FFGraph &DAG, const std::vector<mc::FFVar> &DAGvars, const std::vector<mc::FFVar> &DAGfunctions,
                     const std::vector<babBase::OptimizationVariable> &variables, const unsigned nineqIn,
                     const unsigned neqIn, const unsigned nineqSquashIn, Settings *settingsIn, Logger *loggerIn,
                     std::vector<Constraint> *constraintPropertiesIn, UpperBoundingSolver::UBS_USE useIn)
{

    UBP_SOLVER desiredSolver;
    switch (useIn) {
        case ubp::UpperBoundingSolver::USE_PRE:
            desiredSolver = settingsIn->UBP_solverPreprocessing;
            break;
        case ubp::UpperBoundingSolver::USE_BAB:
            desiredSolver = settingsIn->UBP_solverBab;
            break;
        default:
            throw(MAiNGOException("  Error in UBP factory: unknown intended use for upper bounding solver."));
    }

    switch (desiredSolver) {
        case SOLVER_EVAL: {
            return std::make_shared<UpperBoundingSolver>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn, useIn);
        }
        case SOLVER_COBYLA:    // Fall through to SLSQP
        case SOLVER_BOBYQA:    // Fall through to SLSQP
        case SOLVER_LBFGS:     // Fall through to SLSQP
        case SOLVER_SLSQP: {
            return std::make_shared<UbpNLopt>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn, useIn);
        }
        case SOLVER_IPOPT: {
            return std::make_shared<UbpIpopt>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn, useIn);
        }
        case SOLVER_CLP: {
            return std::make_shared<UbpClp>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn, useIn);
        }
        case SOLVER_KNITRO: {
#ifdef HAVE_KNITRO
            return std::make_shared<UbpKnitro>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn, useIn);
#else
            throw(MAiNGOException("  Error in UbpFactory: Cannot use upper bounding strategy SOLVER_KNITRO: Your MAiNGO build does not contain KNITRO."));
#endif
        }
        case SOLVER_CPLEX: {
#ifdef HAVE_CPLEX
            return std::make_shared<UbpCplex>(DAG, DAGvars, DAGfunctions, variables, nineqIn, neqIn, nineqSquashIn, settingsIn, loggerIn, constraintPropertiesIn, useIn);
#else
            throw(MAiNGOException("  Error in UbpFactory: Cannot use upper bounding strategy SOLVER_CPLEX: Your MAiNGO build does not contain CPLEX."));
#endif
        }
        default:
        {
            std::ostringstream errmsg;
            errmsg << "  Error in UbpFactory: Unknown upper bounding strategy: " << desiredSolver << std::endl;
            throw(MAiNGOException(errmsg.str()));
        }
    }
}